import pytest

import docopt
from docopt import Argument
from docopt import DocoptExit
from docopt import Option
from docopt import Tokens
from docopt import parse_argv


def test_version():
    assert isinstance(docopt.__version__, str)


def test_docopt_ng_more_magic_spellcheck_and_expansion():
    def TS(s):
        return Tokens(s, error=DocoptExit)

    o = [Option("-h"), Option("-v", "--verbose"), Option(None, "--file", 1)]
    assert parse_argv(TS(""), options=o) == []
    assert parse_argv(TS("-h"), options=o) == [Option("-h", None, 0, True)]
    assert parse_argv(TS("-V"), options=o, more_magic=True) == [
        Option("-v", "--verbose", 0, True)
    ]
    assert parse_argv(TS("-h --File f.txt"), options=o, more_magic=True) == [
        Option("-h", None, 0, True),
        Option(None, "--file", 1, "f.txt"),
    ]
    assert parse_argv(TS("-h --fiLe f.txt arg"), options=o, more_magic=True) == [
        Option("-h", None, 0, True),
        Option(None, "--file", 1, "f.txt"),
        Argument(None, "arg"),
    ]
    assert parse_argv(TS("-h -f f.txt arg arg2"), options=o, more_magic=True) == [
        Option("-h", None, 0, True),
        Option(None, "--file", 1, "f.txt"),
        Argument(None, "arg"),
        Argument(None, "arg2"),
    ]


def test_docopt_ng_negative_float():
    args = docopt.docopt(
        "usage: prog --negative_pi=NEGPI NEGTAU", "--negative_pi -3.14 -6.28"
    )
    assert args == {"--negative_pi": "-3.14", "NEGTAU": "-6.28"}


def test_docopt_ng_doubledash_version(capsys: pytest.CaptureFixture):
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        docopt.docopt("usage: prog", version=1, argv="prog --version")
    assert capsys.readouterr().out == "1\n"
    assert pytest_wrapped_e.type == SystemExit


def test_docopt_ng_dot_access_with_dash():
    doc = """Usage: prog [-vqrd] [FILE]
              prog INPUT OUTPUT
              prog --help

    Options:
      -d --dash-arg  test this argument
      -v             print status messages
      -q             report only file names
      -r             show all occurrences of the same error
      --help

    """
    arguments = docopt.docopt(doc, "-v -d file.py")
    assert arguments == {
        "--dash-arg": True,
        "-v": True,
        "-q": False,
        "-r": False,
        "--help": False,
        "FILE": "file.py",
        "INPUT": None,
        "OUTPUT": None,
    }
    assert arguments.v
    assert arguments.FILE == "file.py"
    assert arguments.dash_arg



"""Usage:
  quick_example.py tcp [<host>] [--force] [--timeout=<seconds>]
  quick_example.py serial <port> [--baud=<rate>] [--timeout=<seconds>]
  quick_example.py -h | --help | --version

"""
from docopt import docopt


def load_json_config():
    import json

    # Pretend that we load the following JSON file:
    source = """
        {"--force": true,
         "--timeout": "10",
         "--baud": "9600"}
    """
    return json.loads(source)


def load_ini_config():
    try:  # Python 2
        from ConfigParser import ConfigParser
        from StringIO import StringIO
    except ImportError:  # Python 3
        from configparser import ConfigParser
        from io import StringIO

    # By using `allow_no_value=True` we are allowed to
    # write `--force` instead of `--force=true` below.
    config = ConfigParser(allow_no_value=True)

    # Pretend that we load the following INI file:
    source = """
        [default-arguments]
        --force
        --baud=19200
        <host>=localhost
    """

    # ConfigParser requires a file-like object and
    # no leading whitespace.
    config_file = StringIO("\n".join(source.split()))
    config.readfp(config_file)

    # ConfigParsers sets keys which have no value
    # (like `--force` above) to `None`. Thus we
    # need to substitute all `None` with `True`.
    return dict(
        (key, True if value is None else value)
        for key, value in config.items("default-arguments")
    )


def merge(dict_1, dict_2):
    """Merge two dictionaries.

    Values that evaluate to true take priority over falsy values.
    `dict_1` takes priority over `dict_2`.

    """
    return dict(
        (str(key), dict_1.get(key) or dict_2.get(key))
        for key in set(dict_2) | set(dict_1)
    )


if __name__ == "__main__":
    json_config = load_json_config()
    ini_config = load_ini_config()
    arguments = docopt(__doc__, version="0.1.1rc")

    # Arguments take priority over INI, INI takes priority over JSON:
    result = merge(arguments, merge(ini_config, json_config))

    from pprint import pprint

    print("\nJSON config:")
    pprint(json_config)
    print("\nINI config:")
    pprint(ini_config)
    print("\nResult:")
    pprint(result)



from __future__ import annotations

import re
from textwrap import dedent
from typing import Sequence

import pytest
from pytest import raises

from docopt import Argument
from docopt import Command
from docopt import DocoptExit
from docopt import DocoptLanguageError
from docopt import Either
from docopt import NotRequired
from docopt import OneOrMore
from docopt import Option
from docopt import OptionsShortcut
from docopt import ParsedOptions
from docopt import Required
from docopt import Tokens
from docopt import docopt
from docopt import formal_usage
from docopt import lint_docstring
from docopt import parse_argv
from docopt import parse_docstring_sections
from docopt import parse_longer
from docopt import parse_options
from docopt import parse_pattern
from docopt import parse_shorts
from docopt import transform


def test_pattern_flat():
    assert Required(OneOrMore(Argument("N")), Option("-a"), Argument("M")).flat() == [
        Argument("N"),
        Option("-a"),
        Argument("M"),
    ]
    assert Required(
        NotRequired(OptionsShortcut()), NotRequired(Option("-a", None))
    ).flat(OptionsShortcut) == [OptionsShortcut()]


def test_option():
    assert Option.parse("-h") == Option("-h", None)
    assert Option.parse("--help") == Option(None, "--help")
    assert Option.parse("-h --help") == Option("-h", "--help")
    assert Option.parse("-h, --help") == Option("-h", "--help")

    assert Option.parse("-h TOPIC") == Option("-h", None, 1)
    assert Option.parse("--help TOPIC") == Option(None, "--help", 1)
    assert Option.parse("-h TOPIC --help TOPIC") == Option("-h", "--help", 1)
    assert Option.parse("-h TOPIC, --help TOPIC") == Option("-h", "--help", 1)
    assert Option.parse("-h TOPIC, --help=TOPIC") == Option("-h", "--help", 1)

    assert Option.parse("-h  Description...") == Option("-h", None)
    assert Option.parse("-h --help  Description...") == Option("-h", "--help")
    assert Option.parse("-h TOPIC  Description...") == Option("-h", None, 1)

    assert Option.parse("    -h") == Option("-h", None)

    assert Option.parse("-h TOPIC  Descripton... [default: 2]") == Option(
        "-h", None, 1, "2"
    )
    assert Option.parse("-h TOPIC  Descripton... [default: topic-1]") == Option(
        "-h", None, 1, "topic-1"
    )
    assert Option.parse("--help=TOPIC  ... [default: 3.14]") == Option(
        None, "--help", 1, "3.14"
    )
    assert Option.parse("-h, --help=DIR  ... [default: ./]") == Option(
        "-h", "--help", 1, "./"
    )
    assert Option.parse("-h TOPIC  Descripton... [dEfAuLt: 2]") == Option(
        "-h", None, 1, "2"
    )


def test_option_name():
    assert Option("-h", None).name == "-h"
    assert Option("-h", "--help").name == "--help"
    assert Option(None, "--help").name == "--help"


def test_commands():
    assert docopt("Usage: prog add", "add") == {"add": True}
    assert docopt("Usage: prog [add]", "") == {"add": False}
    assert docopt("Usage: prog [add]", "add") == {"add": True}
    assert docopt("Usage: prog (add|rm)", "add") == {"add": True, "rm": False}
    assert docopt("Usage: prog (add|rm)", "rm") == {"add": False, "rm": True}
    assert docopt("Usage: prog a b", "a b") == {"a": True, "b": True}
    with raises(
        DocoptExit,
        match=r"Warning: found unmatched \(duplicate\?\) arguments.*'b'.*'a'",
    ):
        docopt("Usage: prog a b", "b a")


def test_formal_usage():
    doc = """
    Usage: prog [-hv] ARG
           prog N M

    prog is a program."""
    _, _, usage_body, _ = parse_docstring_sections(doc)
    assert usage_body == " prog [-hv] ARG\n           prog N M\n"
    assert formal_usage(usage_body) == "( [-hv] ARG ) | ( N M )"


def test_parse_argv():
    def TS(s):
        return Tokens(s, error=DocoptExit)

    o = [Option("-h"), Option("-v", "--verbose"), Option("-f", "--file", 1)]
    assert parse_argv(TS(""), options=o) == []
    assert parse_argv(TS("-h"), options=o) == [Option("-h", None, 0, True)]
    assert parse_argv(TS("-h --verbose"), options=o) == [
        Option("-h", None, 0, True),
        Option("-v", "--verbose", 0, True),
    ]
    assert parse_argv(TS("-h --file f.txt"), options=o) == [
        Option("-h", None, 0, True),
        Option("-f", "--file", 1, "f.txt"),
    ]
    assert parse_argv(TS("-h --file f.txt arg"), options=o) == [
        Option("-h", None, 0, True),
        Option("-f", "--file", 1, "f.txt"),
        Argument(None, "arg"),
    ]
    assert parse_argv(TS("-h --file f.txt arg arg2"), options=o) == [
        Option("-h", None, 0, True),
        Option("-f", "--file", 1, "f.txt"),
        Argument(None, "arg"),
        Argument(None, "arg2"),
    ]
    assert parse_argv(TS("-h arg -- -v"), options=o) == [
        Option("-h", None, 0, True),
        Argument(None, "arg"),
        Argument(None, "--"),
        Argument(None, "-v"),
    ]


def test_parse_pattern():
    o = [Option("-h"), Option("-v", "--verbose"), Option("-f", "--file", 1)]
    assert parse_pattern("[ -h ]", options=o) == Required(NotRequired(Option("-h")))
    assert parse_pattern("[ ARG ... ]", options=o) == Required(
        NotRequired(OneOrMore(Argument("ARG")))
    )
    assert parse_pattern("[ -h | -v ]", options=o) == Required(
        NotRequired(Either(Option("-h"), Option("-v", "--verbose")))
    )
    assert parse_pattern("( -h | -v [ --file <f> ] )", options=o) == Required(
        Required(
            Either(
                Option("-h"),
                Required(
                    Option("-v", "--verbose"),
                    NotRequired(Option("-f", "--file", 1, None)),
                ),
            )
        )
    )
    assert parse_pattern("(-h|-v[--file=<f>]N...)", options=o) == Required(
        Required(
            Either(
                Option("-h"),
                Required(
                    Option("-v", "--verbose"),
                    NotRequired(Option("-f", "--file", 1, None)),
                    OneOrMore(Argument("N")),
                ),
            )
        )
    )
    assert parse_pattern("(N [M | (K | L)] | O P)", options=[]) == Required(
        Required(
            Either(
                Required(
                    Argument("N"),
                    NotRequired(
                        Either(
                            Argument("M"),
                            Required(Either(Argument("K"), Argument("L"))),
                        )
                    ),
                ),
                Required(Argument("O"), Argument("P")),
            )
        )
    )
    assert parse_pattern("[ -h ] [N]", options=o) == Required(
        NotRequired(Option("-h")), NotRequired(Argument("N"))
    )
    assert parse_pattern("[options]", options=o) == Required(
        NotRequired(OptionsShortcut())
    )
    assert parse_pattern("[options] A", options=o) == Required(
        NotRequired(OptionsShortcut()), Argument("A")
    )
    assert parse_pattern("-v [options]", options=o) == Required(
        Option("-v", "--verbose"), NotRequired(OptionsShortcut())
    )
    assert parse_pattern("ADD", options=o) == Required(Argument("ADD"))
    assert parse_pattern("<add>", options=o) == Required(Argument("<add>"))
    assert parse_pattern("add", options=o) == Required(Command("add"))


def test_option_match():
    assert Option("-a").match([Option("-a", value=True)]) == (
        True,
        [],
        [Option("-a", value=True)],
    )
    assert Option("-a").match([Option("-x")]) == (False, [Option("-x")], [])
    assert Option("-a").match([Argument("N")]) == (False, [Argument("N")], [])
    assert Option("-a").match([Option("-x"), Option("-a"), Argument("N")]) == (
        True,
        [Option("-x"), Argument("N")],
        [Option("-a")],
    )
    assert Option("-a").match([Option("-a", value=True), Option("-a")]) == (
        True,
        [Option("-a")],
        [Option("-a", value=True)],
    )


def test_argument_match():
    assert Argument("N").match([Argument(None, 9)]) == (True, [], [Argument("N", 9)])
    assert Argument("N").match([Option("-x")]) == (False, [Option("-x")], [])
    assert Argument("N").match([Option("-x"), Option("-a"), Argument(None, 5)]) == (
        True,
        [Option("-x"), Option("-a")],
        [Argument("N", 5)],
    )
    assert Argument("N").match([Argument(None, 9), Argument(None, 0)]) == (
        True,
        [Argument(None, 0)],
        [Argument("N", 9)],
    )


def test_command_match():
    assert Command("c").match([Argument(None, "c")]) == (True, [], [Command("c", True)])
    assert Command("c").match([Option("-x")]) == (False, [Option("-x")], [])
    assert Command("c").match([Option("-x"), Option("-a"), Argument(None, "c")]) == (
        True,
        [Option("-x"), Option("-a")],
        [Command("c", True)],
    )
    assert Either(Command("add", False), Command("rm", False)).match(
        [Argument(None, "rm")]
    ) == (True, [], [Command("rm", True)])


def test_optional_match():
    assert NotRequired(Option("-a")).match([Option("-a")]) == (True, [], [Option("-a")])
    assert NotRequired(Option("-a")).match([]) == (True, [], [])
    assert NotRequired(Option("-a")).match([Option("-x")]) == (True, [Option("-x")], [])
    assert NotRequired(Option("-a"), Option("-b")).match([Option("-a")]) == (
        True,
        [],
        [Option("-a")],
    )
    assert NotRequired(Option("-a"), Option("-b")).match([Option("-b")]) == (
        True,
        [],
        [Option("-b")],
    )
    assert NotRequired(Option("-a"), Option("-b")).match([Option("-x")]) == (
        True,
        [Option("-x")],
        [],
    )
    assert NotRequired(Argument("N")).match([Argument(None, 9)]) == (
        True,
        [],
        [Argument("N", 9)],
    )
    assert NotRequired(Option("-a"), Option("-b")).match(
        [Option("-b"), Option("-x"), Option("-a")]
    ) == (True, [Option("-x")], [Option("-a"), Option("-b")])


def test_required_match():
    assert Required(Option("-a")).match([Option("-a")]) == (True, [], [Option("-a")])
    assert Required(Option("-a")).match([]) == (False, [], [])
    assert Required(Option("-a")).match([Option("-x")]) == (False, [Option("-x")], [])
    assert Required(Option("-a"), Option("-b")).match([Option("-a")]) == (
        False,
        [Option("-a")],
        [],
    )


def test_either_match():
    assert Either(Option("-a"), Option("-b")).match([Option("-a")]) == (
        True,
        [],
        [Option("-a")],
    )
    assert Either(Option("-a"), Option("-b")).match([Option("-a"), Option("-b")]) == (
        True,
        [Option("-b")],
        [Option("-a")],
    )
    assert Either(Option("-a"), Option("-b")).match([Option("-x")]) == (
        False,
        [Option("-x")],
        [],
    )
    assert Either(Option("-a"), Option("-b"), Option("-c")).match(
        [Option("-x"), Option("-b")]
    ) == (True, [Option("-x")], [Option("-b")])
    assert Either(Argument("M"), Required(Argument("N"), Argument("M"))).match(
        [Argument(None, 1), Argument(None, 2)]
    ) == (
        True,
        [],
        [Argument("N", 1), Argument("M", 2)],
    )


def test_one_or_more_match():
    assert OneOrMore(Argument("N")).match([Argument(None, 9)]) == (
        True,
        [],
        [Argument("N", 9)],
    )
    assert OneOrMore(Argument("N")).match([]) == (False, [], [])
    assert OneOrMore(Argument("N")).match([Option("-x")]) == (False, [Option("-x")], [])
    assert OneOrMore(Argument("N")).match([Argument(None, 9), Argument(None, 8)]) == (
        True,
        [],
        [Argument("N", 9), Argument("N", 8)],
    )
    assert OneOrMore(Argument("N")).match(
        [Argument(None, 9), Option("-x"), Argument(None, 8)]
    ) == (True, [Option("-x")], [Argument("N", 9), Argument("N", 8)])
    assert OneOrMore(Option("-a")).match(
        [Option("-a"), Argument(None, 8), Option("-a")]
    ) == (True, [Argument(None, 8)], [Option("-a"), Option("-a")])
    assert OneOrMore(Option("-a")).match([Argument(None, 8), Option("-x")]) == (
        False,
        [Argument(None, 8), Option("-x")],
        [],
    )
    assert OneOrMore(Required(Option("-a"), Argument("N"))).match(
        [Option("-a"), Argument(None, 1), Option("-x"), Option("-a"), Argument(None, 2)]
    ) == (
        True,
        [Option("-x")],
        [Option("-a"), Argument("N", 1), Option("-a"), Argument("N", 2)],
    )
    assert OneOrMore(NotRequired(Argument("N"))).match([Argument(None, 9)]) == (
        True,
        [],
        [Argument("N", 9)],
    )


def test_list_argument_match():
    assert Required(Argument("N"), Argument("N")).fix().match(
        [Argument(None, "1"), Argument(None, "2")]
    ) == (True, [], [Argument("N", ["1", "2"])])
    assert OneOrMore(Argument("N")).fix().match(
        [Argument(None, "1"), Argument(None, "2"), Argument(None, "3")]
    ) == (True, [], [Argument("N", ["1", "2", "3"])])
    assert Required(Argument("N"), OneOrMore(Argument("N"))).fix().match(
        [Argument(None, "1"), Argument(None, "2"), Argument(None, "3")]
    ) == (
        True,
        [],
        [Argument("N", ["1", "2", "3"])],
    )
    assert Required(Argument("N"), Required(Argument("N"))).fix().match(
        [Argument(None, "1"), Argument(None, "2")]
    ) == (True, [], [Argument("N", ["1", "2"])])


def test_basic_pattern_matching():
    # ( -a N [ -x Z ] )
    pattern = Required(
        Option("-a"), Argument("N"), NotRequired(Option("-x"), Argument("Z"))
    )
    # -a N
    assert pattern.match([Option("-a"), Argument(None, 9)]) == (
        True,
        [],
        [Option("-a"), Argument("N", 9)],
    )
    # -a -x N Z
    assert pattern.match(
        [Option("-a"), Option("-x"), Argument(None, 9), Argument(None, 5)]
    ) == (
        True,
        [],
        [Option("-a"), Argument("N", 9), Option("-x"), Argument("Z", 5)],
    )
    # -x N Z  # BZZ!
    assert pattern.match([Option("-x"), Argument(None, 9), Argument(None, 5)]) == (
        False,
        [Option("-x"), Argument(None, 9), Argument(None, 5)],
        [],
    )


def test_pattern_either():
    assert transform(Option("-a")) == Either(Required(Option("-a")))
    assert transform(Argument("A")) == Either(Required(Argument("A")))
    assert transform(
        Required(Either(Option("-a"), Option("-b")), Option("-c"))
    ) == Either(
        Required(Option("-a"), Option("-c")), Required(Option("-b"), Option("-c"))
    )
    assert transform(
        NotRequired(Option("-a"), Either(Option("-b"), Option("-c")))
    ) == Either(
        Required(Option("-b"), Option("-a")), Required(Option("-c"), Option("-a"))
    )
    assert transform(
        Either(Option("-x"), Either(Option("-y"), Option("-z")))
    ) == Either(Required(Option("-x")), Required(Option("-y")), Required(Option("-z")))
    assert transform(OneOrMore(Argument("N"), Argument("M"))) == Either(
        Required(Argument("N"), Argument("M"), Argument("N"), Argument("M"))
    )


def test_pattern_fix_repeating_arguments():
    assert Required(Option("-a")).fix_repeating_arguments() == Required(Option("-a"))
    assert Required(Argument("N", None)).fix_repeating_arguments() == Required(
        Argument("N", None)
    )
    assert Required(Argument("N"), Argument("N")).fix_repeating_arguments() == Required(
        Argument("N", []), Argument("N", [])
    )
    assert Either(Argument("N"), OneOrMore(Argument("N"))).fix() == Either(
        Argument("N", []), OneOrMore(Argument("N", []))
    )


def test_set():
    assert Argument("N") == Argument("N")
    assert set([Argument("N"), Argument("N")]) == set([Argument("N")])


def test_pattern_fix_identities_1():
    pattern = Required(Argument("N"), Argument("N"))
    assert pattern.children[0] == pattern.children[1]
    assert pattern.children[0] is not pattern.children[1]
    pattern.fix_identities()
    assert pattern.children[0] is pattern.children[1]


def test_pattern_fix_identities_2():
    pattern = Required(NotRequired(Argument("X"), Argument("N")), Argument("N"))
    assert pattern.children[0].children[1] == pattern.children[1]
    assert pattern.children[0].children[1] is not pattern.children[1]
    pattern.fix_identities()
    assert pattern.children[0].children[1] is pattern.children[1]


@pytest.mark.parametrize("tokens", [[], ["not_a_long_option"]])
def test_parse_longer__rejects_inappropriate_token(tokens: list[str]):
    with raises(
        ValueError, match=r"parse_longer got what appears to be an invalid token"
    ):
        parse_longer(Tokens(tokens), [])


def test_parse_longer__rejects_duplicate_long_options():
    options = [Option(None, "--foo"), Option(None, "--foo")]
    with raises(DocoptLanguageError, match=r"foo is not a unique prefix"):
        parse_longer(Tokens("--foo"), options)


@pytest.mark.parametrize("tokens", [[], ["not_a_short_option"]])
def test_parse_shorts__rejects_inappropriate_token(tokens: list[str]):
    with raises(
        ValueError, match=r"parse_shorts got what appears to be an invalid token"
    ):
        parse_shorts(Tokens(tokens), [])


def test_parse_shorts__rejects_duplicate_short_options():
    options = [Option("-f"), Option("-f")]
    with raises(DocoptLanguageError, match=r"-f is specified ambiguously 2 times"):
        parse_shorts(Tokens("-f"), options)


def test_long_options_error_handling():
    #    with raises(DocoptLanguageError):
    #        docopt('Usage: prog --non-existent', '--non-existent')
    #    with raises(DocoptLanguageError):
    #        docopt('Usage: prog --non-existent')
    with raises(
        DocoptExit,
        match=r"Warning: found unmatched \(duplicate\?\) arguments.*--non-existent",
    ):
        docopt("Usage: prog", "--non-existent")
    with raises(
        DocoptExit, match=r"Warning: found unmatched \(duplicate\?\) arguments.*--ver\b"
    ):
        docopt(
            "Usage: prog [--version --verbose]\n" "Options: --version\n --verbose",
            "--ver",
        )
    # --long is missing ARG in usage
    with raises(DocoptLanguageError, match=r"unmatched '\('"):
        docopt("Usage: prog --long\nOptions: --long ARG")
    with raises(DocoptExit, match=r"--long requires argument"):
        docopt("Usage: prog --long ARG\nOptions: --long ARG", "--long")
    with raises(DocoptLanguageError, match=r"--long must not have an argument"):
        docopt("Usage: prog --long=ARG\nOptions: --long")
    with raises(DocoptExit, match=r"--long must not have an argument"):
        docopt("Usage: prog --long\nOptions: --long", "--long=ARG")


def test_short_options_error_handling():
    with raises(DocoptLanguageError, match=r"-x is specified ambiguously 2 times"):
        docopt("Usage: prog -x\nOptions: -x  this\n -x  that")

    with raises(
        DocoptExit, match=r"Warning: found unmatched \(duplicate\?\) arguments.*-x"
    ):
        docopt("Usage: prog", "-x")

    with raises(DocoptLanguageError):
        docopt("Usage: prog -o\nOptions: -o ARG")
    with raises(DocoptExit, match=r"-o requires argument"):
        docopt("Usage: prog -o ARG\nOptions: -o ARG", "-o")


def test_matching_paren():
    with raises(DocoptLanguageError, match=r"unmatched '\['"):
        docopt("Usage: prog [a [b]")
    with raises(DocoptLanguageError, match=r"unexpected ending: '\)'"):
        docopt("Usage: prog [a [b] ] c )")


def test_allow_double_dash():
    assert docopt("usage: prog [-o] [--] <arg>\nkptions: -o", "-- -o") == {
        "-o": False,
        "<arg>": "-o",
        "--": True,
    }
    assert docopt("usage: prog [-o] [--] <arg>\nkptions: -o", "-o 1") == {
        "-o": True,
        "<arg>": "1",
        "--": False,
    }
    with raises(
        DocoptExit, match=r"Warning: found unmatched \(duplicate\?\) arguments.*-o\b"
    ):  # "--" is not allowed; FIXME?
        docopt("usage: prog [-o] <arg>\noptions:-o", "-- -o")


def test_docopt(capsys: pytest.CaptureFixture):
    doc = """Usage: prog [-v] A

             Options: -v  Be verbose."""
    assert docopt(doc, "arg") == {"-v": False, "A": "arg"}
    assert docopt(doc, "-v arg") == {"-v": True, "A": "arg"}

    doc = """Usage: prog [-vqr] [FILE]
              prog INPUT OUTPUT
              prog --help

    Options:
      -v  print status messages
      -q  report only file names
      -r  show all occurrences of the same error
      --help

    """
    a = docopt(doc, "-v file.py")
    assert a == {
        "-v": True,
        "-q": False,
        "-r": False,
        "--help": False,
        "FILE": "file.py",
        "INPUT": None,
        "OUTPUT": None,
    }

    a = docopt(doc, "-v")
    assert a == {
        "-v": True,
        "-q": False,
        "-r": False,
        "--help": False,
        "FILE": None,
        "INPUT": None,
        "OUTPUT": None,
    }

    with raises(
        DocoptExit,
        match=r"Warning: found unmatched \(duplicate\?\) arguments.*output\.py",
    ):
        docopt(doc, "-v input.py output.py")

    with raises(
        DocoptExit,
        match=r"Warning: found unmatched \(duplicate\?\) arguments.*--fake",
    ):
        docopt(doc, "--fake")

    capsys.readouterr()  # clear any captured output

    with raises(SystemExit):
        docopt(doc, "--hel")
    assert capsys.readouterr().out.startswith("Usage: prog")


@pytest.mark.parametrize(
    "items, expected",
    [
        ({}, "{}"),
        (
            {"<z>": None, "<a>": None, "--foo": "abc", "--bar": True},
            ("{'--bar': True,\n '--foo': 'abc',\n '<a>': None,\n '<z>': None}"),
        ),
    ],
)
def test_docopt_result_dict_repr(items: dict[str, object], expected: str):
    assert repr(ParsedOptions(items)) == expected


@pytest.mark.parametrize(
    "args, before_usage_val", [("", None), ("--before-usage=2", "2")]
)
def test_docopt__usage_descriptions_cant_bridge_usage_section(
    args: str, before_usage_val: str | None
):
    # For compatibility with docopt 0.6.2 we support option descriptions
    # before the usage and after (but not inside usage). However, a
    # description cannot start in one part and continue in the next.
    # i.e. the default value after Usage does not apply to
    # --before-usage
    usage = """\
My prog

--before-usage VAL

Usage:
    prog [options]

[default: 42]
Options:
    --after-usage
"""
    assert docopt(usage, args) == {
        "--before-usage": before_usage_val,
        "--after-usage": False,
    }


def test_language_errors():
    with raises(
        DocoptLanguageError,
        match=r'Failed to parse doc: "usage:" section \(case-insensitive\) not '
        r"found\. Check http://docopt\.org/ for examples of how your doc "
        r"should look\.",
    ):
        docopt("no usage with colon here")
    with raises(
        DocoptLanguageError, match=r'More than one "usage:" \(case-insensitive\)'
    ):
        docopt("usage: here \n\n and again usage: here")


def test_issue_40(capsys: pytest.CaptureFixture):
    with raises(SystemExit):  # i.e. shows help
        docopt("usage: prog --help-commands | --help", "--help")
    assert capsys.readouterr().out.startswith("usage: prog --help-commands | --help")

    assert docopt("usage: prog --aabb | --aa", "--aa") == {
        "--aabb": False,
        "--aa": True,
    }


def test_count_multiple_flags():
    assert docopt("usage: prog [-v]", "-v") == {"-v": True}
    assert docopt("usage: prog [-vv]", "") == {"-v": 0}
    assert docopt("usage: prog [-vv]", "-v") == {"-v": 1}
    assert docopt("usage: prog [-vv]", "-vv") == {"-v": 2}
    with raises(
        DocoptExit, match=r"Warning: found unmatched \(duplicate\?\) arguments.*'-v'"
    ):
        docopt("usage: prog [-vv]", "-vvv")
    assert docopt("usage: prog [-v | -vv | -vvv]", "-vvv") == {"-v": 3}
    assert docopt("usage: prog -v...", "-vvvvvv") == {"-v": 6}
    assert docopt("usage: prog [--ver --ver]", "--ver --ver") == {"--ver": 2}


def test_any_options_parameter():
    with raises(
        DocoptExit,
        match=r"Warning: found unmatched \(duplicate\?\) arguments"
        r".*-f.*-o.*-o.*--bar.*--spam.*eggs",
    ):
        docopt("usage: prog [options]", "-foo --bar --spam=eggs")
    #    assert docopt('usage: prog [options]', '-foo --bar --spam=eggs',
    #                  any_options=True) == {'-f': True, '-o': 2,
    #                                         '--bar': True, '--spam': 'eggs'}
    with raises(
        DocoptExit,
        match=r"Warning: found unmatched \(duplicate\?\) arguments"
        r".*--foo.*--bar.*--bar",
    ):
        docopt("usage: prog [options]", "--foo --bar --bar")
    #    assert docopt('usage: prog [options]', '--foo --bar --bar',
    #                  any_options=True) == {'--foo': True, '--bar': 2}
    with raises(
        DocoptExit,
        match=r"Warning: found unmatched \(duplicate\?\) arguments"
        r".*--bar.*--bar.*--bar.*-f.*-f.*-f.*-f",
    ):
        docopt("usage: prog [options]", "--bar --bar --bar -ffff")
    #    assert docopt('usage: prog [options]', '--bar --bar --bar -ffff',
    #                  any_options=True) == {'--bar': 3, '-f': 4}
    with raises(
        DocoptExit,
        match=r"Warning: found unmatched \(duplicate\?\) arguments"
        r".*--long.*arg.*--long.*another",
    ):
        docopt("usage: prog [options]", "--long=arg --long=another")


#    assert docopt('usage: prog [options]', '--long=arg --long=another',
#                  any_options=True) == {'--long': ['arg', 'another']}


# def test_options_shortcut_multiple_commands():
#    # any_options is disabled
#    assert docopt('usage: prog c1 [options] prog c2 [options]',
#        'c2 -o', any_options=True) == {'-o': True, 'c1': False, 'c2': True}
#    assert docopt('usage: prog c1 [options] prog c2 [options]',
#        'c1 -o', any_options=True) == {'-o': True, 'c1': True, 'c2': False}


def test_default_value_for_positional_arguments():
    doc = """Usage: prog [--data=<data>...]\n
             Options:\n\t-d --data=<arg>    Input data [default: x]
          """
    a = docopt(doc, "")
    assert a == {"--data": ["x"]}
    doc = """Usage: prog [--data=<data>...]\n
             Options:\n\t-d --data=<arg>    Input data [default: x y]
          """
    a = docopt(doc, "")
    assert a == {"--data": ["x", "y"]}
    doc = """Usage: prog [--data=<data>...]\n
             Options:\n\t-d --data=<arg>    Input data [default: x y]
          """
    a = docopt(doc, "--data=this")
    assert a == {"--data": ["this"]}


def test_issue_59():
    assert docopt("usage: prog --long=<a>", "--long=") == {"--long": ""}
    assert docopt("usage: prog -l <a>\n" "options: -l <a>", ["-l", ""]) == {"-l": ""}


def test_options_first():
    assert docopt("usage: prog [--opt] [<args>...]", "--opt this that") == {
        "--opt": True,
        "<args>": ["this", "that"],
    }
    assert docopt("usage: prog [--opt] [<args>...]", "this that --opt") == {
        "--opt": True,
        "<args>": ["this", "that"],
    }
    assert docopt(
        "usage: prog [--opt] [<args>...]", "this that --opt", options_first=True
    ) == {"--opt": False, "<args>": ["this", "that", "--opt"]}


def test_issue_68_options_shortcut_does_not_include_options_in_usage_pattern():
    args = docopt("usage: prog [-ab] [options]\n" "options: -x\n -y", "-ax")
    # Need to use `is` (not `==`) since we want to make sure
    # that they are not 1/0, but strictly True/False:
    assert args["-a"] is True
    assert args["-b"] is False
    assert args["-x"] is True
    assert args["-y"] is False


def test_issue_65_evaluate_argv_when_called_not_when_imported():
    import sys

    sys.argv = "prog -a".split()
    assert docopt("usage: prog [-ab]") == {"-a": True, "-b": False}
    sys.argv = "prog -b".split()
    assert docopt("usage: prog [-ab]") == {"-a": False, "-b": True}


def test_issue_71_double_dash_is_not_a_valid_option_argument():
    with raises(DocoptExit, match=r"--log requires argument"):
        docopt("usage: prog [--log=LEVEL] [--] <args>...", "--log -- 1 2")
    with raises(DocoptExit, match=r"-l requires argument"):
        docopt("usage: prog [-l LEVEL] [--] <args>...\noptions: -l LEVEL", "-l -- 1 2")


option_examples: Sequence[tuple[str, Sequence[Option]]] = [
    ("", []),
    ("Some content\nbefore the first option.", []),
    ("-f", [Option("-f", None, 0, False)]),
    ("-f  Description.", [Option("-f", None, 0, False)]),
    ("-f ARG  Description.", [Option("-f", None, 1, None)]),
    ("-f ARG  Description. [default: 42]", [Option("-f", None, 1, "42")]),
    ("--foo", [Option(None, "--foo", 0, False)]),
    ("--foo  Description.", [Option(None, "--foo", 0, False)]),
    ("--foo ARG  Description.", [Option(None, "--foo", 1, None)]),
    ("--foo ARG  Description. [default: 42]", [Option(None, "--foo", 1, "42")]),
    # Options can wrap over multiple lines
    (
        """\
         \t --foo ARG, -f ARG  With a long

         wrapped description
           \t [default: 42]
         """,
        [Option("-f", "--foo", 1, "42")],
    ),
    # Options can start after whitespace
    (
        "\t--foo=<arg>  [default: bar]",
        [Option(None, "--foo", 1, "bar")],
    ),
    (
        " \t -f ARG, --foo ARG  Description. [default: 42]",
        [Option("-f", "--foo", 1, "42")],
    ),
    # Options can start on the same line as an "options:" heading
    (
        "options:-f ARG, --foo ARG  Description. [default: 42]",
        [Option("-f", "--foo", 1, "42")],
    ),
    (
        "  Special oPtioNs:  --foo ARG  Description. [default: 42]",
        [Option(None, "--foo", 1, "42")],
    ),
    (
        "  other options: --foo ARG  Description. [default: 42]",
        [Option(None, "--foo", 1, "42")],
    ),
    (
        """\
        -a  This is the first option

            -b=<x>  Options don't have to be in an options section

        Options:
            -c, --charlie  This describes the option.
            --delta, -d
                This option has the desc on another line.

        --echo   This option starts after a blank line.

            -f --foxtrot   This option has no comma

        Other Options:
            -g VAL     This option is after another section heading.
                       [default: gval]
        options:-h  This option is on the same line as a heading
        oPtioNs:--india
              oPtIons:  -j X

                        [default: jval]
            and more Options:  --k X  [default: kval]
        """,
        [
            Option("-a", None, 0, False),
            Option("-b", None, 1, None),
            Option("-c", "--charlie", 0, False),
            Option("-d", "--delta", 0, False),
            Option(None, "--echo", 0, False),
            Option("-f", "--foxtrot", 0, False),
            Option("-g", None, 1, "gval"),
            Option("-h", None, 0, False),
            Option(None, "--india", 0, False),
            Option("-j", None, 1, "jval"),
            Option(None, "--k", 1, "kval"),
        ],
    ),
    # Option with description (or other content) on following line.
    (
        """
        Options:
            -a
            -b
         description of b
            -c
        Other Options:
            -d
        Other Options:-e
        """,
        [
            Option("-a", None, 0, False),
            Option("-b", None, 0, False),
            Option("-c", None, 0, False),
            Option("-d", None, 0, False),
            Option("-e", None, 0, False),
        ],
    ),
    # Option-like things which aren't actually options
    (
        """
        --option1 <x>  This really is an option.
                       And it has a default [default: 42]

        Talking about options:
            Here we're talking about options and defaults, like [default: 3] and
            options such as --foo, but we're not intending to define them. And
            although the default of 3 I just mentioned does not get picked up as
            the default of --option1, defined above.

            But if we happen to start a line of our prose with an option, like
            -b then we are unfortunately defining an option. And "then" acts as
            an argument for -b, so it accepts an argument.

            Options are also allowed to start on the same line as an option
            heading, so this is an option:
            options: --option2

            And this also works after several words, so options: --option3  is
            also an option. But options after other heading-like things aren't
            picked up, so this isn't an option:
            things: --not-an-option

        -o, --option4 <x>  This is also a real option
        """,
        [
            Option(None, "--option1", 1, "42"),
            Option("-b", None, 1, None),
            Option(None, "--option2", 0, False),
            Option(None, "--option3", 0, False),
            Option("-o", "--option4", 1, None),
        ],
    ),
]
option_examples = [(dedent(doc), options) for (doc, options) in option_examples]


@pytest.mark.parametrize("descriptions, options", option_examples)
def test_parse_options(descriptions, options):
    assert parse_options(descriptions) == options


@pytest.mark.parametrize(
    "before",
    [
        pytest.param("", id="empty"),
        pytest.param("This is a prog\n", id="1line"),
        pytest.param(
            "This is a prog\n\nInfo:\n Blah blah\n\n"
            # contains usage: but not a usage section
            "Ingredients in pork sausage:\nBlah blah\n",
            id="preceding_sections",
        ),
    ],
)
@pytest.mark.parametrize(
    "header",
    [
        pytest.param("usage:", id="simple"),
        pytest.param("uSaGe:", id="odd_case"),
        pytest.param("My Program's Usage:", id="long"),
        pytest.param("  Indented Usage:", id="indented"),
    ],
)
@pytest.mark.parametrize(
    "body",
    [
        pytest.param("prog [options]", id="simple"),
        pytest.param(" prog [options]", id="space_simple"),
        pytest.param("\tprog [options]", id="tab_simple"),
        pytest.param(" \t prog [options]", id="WS_simple"),
        pytest.param("\n prog [options]", id="LF_simple"),
        pytest.param("\n prog [options]\n", id="LF_simple_LF"),
        pytest.param("prog [options] cmd1\n prog [options] cmd2\n", id="multiple_LF"),
        pytest.param("\n prog [options] cmd1\n prog [options] cmd2", id="LF_multiple"),
        pytest.param(
            "\n prog [options] cmd1\n prog [options] cmd2\n", id="LF_multiple_LF"
        ),
        pytest.param(
            """\
 prog [options] cmd1
   [--foo --bar]
   [--baz --boz]
 prog [options] cmd2
""",
            id="wrapped_arguments",
        ),
    ],
)
@pytest.mark.parametrize(
    "after",
    [
        pytest.param("", id="empty"),
        pytest.param("This can be\nany content.\n", id="text"),
        pytest.param("Options: -a  All", id="single_line"),
    ],
)
def test_parse_docstring_sections(before: str, header: str, body: str, after: str):
    if after and not body.endswith("\n"):
        body = body + "\n"
    assert parse_docstring_sections(before + header + body + after) == (
        (before, header, body, after)
    )


@pytest.mark.parametrize(
    "invalid_docstring",
    [
        pytest.param("", id="empty"),
        pytest.param(
            """\
            This doc has no usage heading

                myprog [options]

            Options:
                --foo
                --bar
            """,
            id="no_usage_heading",
        ),
    ],
)
def test_parse_docstring_sections__reports_invalid_docstrings(invalid_docstring: str):
    with pytest.raises(
        DocoptLanguageError,
        match=re.escape(
            'Failed to parse doc: "usage:" section (case-insensitive) not found'
        ),
    ):
        parse_docstring_sections(dedent(invalid_docstring))


@pytest.mark.parametrize(
    "doc, error_message",
    [
        pytest.param(
            """\
            My prog.

            Usage:
                myprog [options]
                Options:
                    --foo
                    --bar
            """,
            'Failed to parse docstring: "options:" (case-insensitive) was '
            'found in "usage:" section.',
            id="options_in_usage",
        ),
        pytest.param(
            """\
            My prog.

            Usage:
                myprog [options]

            More Usage:
                Blah blah.
            """,
            'Failed to parse docstring: More than one "usage:" '
            "(case-insensitive) section found.",
            id="multiple_usage_sections",
        ),
        pytest.param(
            """\
            This docstring has nothing in its usage.

            Usage:""",
            'Failed to parse docstring: "usage:" section is empty.',
            id="empty_usage_section",
        ),
        pytest.param(
            """\
            This docstring has only whitespace in its usage.

            Usage:

            Options:""",
            'Failed to parse docstring: "usage:" section is empty.',
            id="whitespace_usage_section",
        ),
    ],
)
def test_lint_docstring(doc: str, error_message: str):
    doc_sections = parse_docstring_sections(dedent(doc))
    with pytest.raises(DocoptLanguageError, match=re.escape(error_message)):
        lint_docstring(doc_sections)



#!/usr/bin/env python
"""
This example uses docopt with the built in cmd module to demonstrate an
interactive command application.

Usage:
    my_program tcp <host> <port> [--timeout=<seconds>]
    my_program serial <port> [--baud=<n>] [--timeout=<seconds>]
    my_program (-i | --interactive)
    my_program (-h | --help | --version)

Options:
    -i, --interactive  Interactive Mode
    -h, --help  Show this screen and exit.
    --baud=<n>  Baudrate [default: 9600]
"""

import cmd
import sys

from docopt import DocoptExit
from docopt import docopt


def docopt_cmd(func):
    """
    This decorator is used to simplify the try/except block and pass the result
    of the docopt parsing to the called action.
    """

    def fn(self, arg):
        try:
            opt = docopt(fn.__doc__, arg)

        except DocoptExit as e:
            # The DocoptExit is thrown when the args do not match.
            # We print a message to the user and the usage block.

            print("Invalid Command!")
            print(e)
            return

        except SystemExit:
            # The SystemExit exception prints the usage for --help
            # We do not need to do the print here.

            return

        return func(self, opt)

    fn.__name__ = func.__name__
    fn.__doc__ = func.__doc__
    fn.__dict__.update(func.__dict__)
    return fn


class MyInteractive(cmd.Cmd):
    intro = "Welcome to my interactive program! (type help for a list of commands.)"
    prompt = "(my_program) "
    file = None

    @docopt_cmd
    def do_tcp(self, arg):
        """Usage: tcp <host> <port> [--timeout=<seconds>]"""

        print(arg)

    @docopt_cmd
    def do_serial(self, arg):
        """Usage: serial <port> [--baud=<n>] [--timeout=<seconds>]

        Options:
            --baud=<n>  Baudrate [default: 9600]
        """

        print(arg)

    def do_quit(self, arg):
        """Quits out of Interactive Mode."""

        print("Good Bye!")
        exit()


opt = docopt(__doc__, sys.argv[1:])

if opt["--interactive"]:
    MyInteractive().cmdloop()

print(opt)



"""usage: git clone [options] [--] <repo> [<dir>]

    -v, --verbose         be more verbose
    -q, --quiet           be more quiet
    --progress            force progress reporting
    -n, --no-checkout     don't create a checkout
    --bare                create a bare repository
    --mirror              create a mirror repository (implies bare)
    -l, --local           to clone from a local repository
    --no-hardlinks        don't use local hardlinks, always copy
    -s, --shared          setup as shared repository
    --recursive           initialize submodules in the clone
    --recurse-submodules  initialize submodules in the clone
    --template <template-directory>
                          directory from which templates will be used
    --reference <repo>    reference repository
    -o, --origin <branch>
                          use <branch> instead of 'origin' to track upstream
    -b, --branch <branch>
                          checkout <branch> instead of the remote's HEAD
    -u, --upload-pack <path>
                          path to git-upload-pack on the remote
    --depth <depth>       create a shallow clone of that depth

"""
from docopt import docopt

if __name__ == "__main__":
    print(docopt(__doc__))



import json
import re
from pathlib import Path
from typing import Generator
from typing import Sequence
from unittest import mock

import pytest

import docopt


def pytest_collect_file(file_path: Path, path, parent):
    if file_path.suffix == ".docopt" and file_path.stem.startswith("test"):
        return DocoptTestFile.from_parent(path=file_path, parent=parent)


def parse_test(raw: str):
    raw = re.compile("#.*$", re.M).sub("", raw).strip()
    if raw.startswith('"""'):
        raw = raw[3:]

    for i, fixture in enumerate(raw.split('r"""')):
        if i == 0:
            if not fixture.strip() == "":
                raise DocoptTestException(
                    f"Unexpected content before first testcase: {fixture}"
                )
            continue

        try:
            doc, _, body = fixture.partition('"""')
            cases = []
            for case in body.split("$")[1:]:
                argv, _, expect = case.strip().partition("\n")
                try:
                    expect = json.loads(expect)
                except json.JSONDecodeError as e:
                    raise DocoptTestException(
                        f"The test case JSON is invalid: {expect!r} - {e}."
                    )
                prog, _, argv = argv.strip().partition(" ")
                cases.append((prog, argv, expect))
            if len(cases) == 0:
                raise DocoptTestException(
                    "No test cases follow the doc. Each example must have at "
                    "least one test case starting with '$'"
                )
        except Exception as e:
            raise DocoptTestException(
                f"Failed to parse test case {i}. {e}\n"
                f'The test\'s definition is:\nr"""{fixture}'
            ) from None
        yield doc, cases


class DocoptTestFile(pytest.File):
    def collect(self):
        raw = self.path.open().read()
        for i, (doc, cases) in enumerate(parse_test(raw), 1):
            name = f"{self.path.stem}({i})"
            for case in cases:
                yield DocoptTestItem.from_parent(
                    name=name, parent=self, doc=doc, case=case
                )


class DocoptTestItem(pytest.Item):
    def __init__(self, name, parent, doc, case):
        super(DocoptTestItem, self).__init__(name, parent)
        self.doc = doc
        self.prog, self.argv, self.expect = case

    def runtest(self):
        try:
            result = docopt.docopt(self.doc, argv=self.argv)
        except docopt.DocoptExit:
            result = "user-error"

        if self.expect != result:
            raise DocoptTestException(self, result)

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        if isinstance(excinfo.value, DocoptTestException):
            return "\n".join(
                (
                    "usecase execution failed:",
                    self.doc.rstrip(),
                    f"$ {self.prog} {self.argv}",
                    f"result> {json.dumps(excinfo.value.args[1])}",
                    f"expect> {json.dumps(self.expect)}",
                )
            )
        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.path, 0, f"usecase: {self.name}"


class DocoptTestException(Exception):
    pass


@pytest.fixture(autouse=True)
def override_sys_argv(argv: Sequence[str]) -> Generator[None, None, None]:
    """Patch `sys.argv` with a fixed value during tests.

    A lot of docopt tests call docopt() without specifying argv, which uses
    `sys.argv` by default, so a predictable value for it is necessary.
    """
    with mock.patch("sys.argv", new=argv):
        yield


@pytest.fixture
def argv() -> Sequence[str]:
    """The `sys.argv` value seen inside tests."""
    return ["exampleprogram"]



"""usage: git push [options] [<repository> [<refspec>...]]

    -h, --help
    -v, --verbose         be more verbose
    -q, --quiet           be more quiet
    --repo <repository>   repository
    --all                 push all refs
    --mirror              mirror all refs
    --delete              delete refs
    --tags                push tags (can't be used with --all or --mirror)
    -n, --dry-run         dry run
    --porcelain           machine-readable output
    -f, --force           force updates
    --thin                use thin pack
    --receive-pack <receive-pack>
                          receive pack program
    --exec <receive-pack>
                          receive pack program
    -u, --set-upstream    set upstream for git pull/status
    --progress            force progress reporting

"""
from docopt import docopt

if __name__ == "__main__":
    print(docopt(__doc__))



#! /usr/bin/env python
"""
usage: git [--version] [--exec-path=<path>] [--html-path]
           [-p|--paginate|--no-pager] [--no-replace-objects]
           [--bare] [--git-dir=<path>] [--work-tree=<path>]
           [-c <name>=<value>] [--help]
           <command> [<args>...]

options:
   -c <name=value>
   -h, --help
   -p, --paginate

The most commonly used git commands are:
   add        Add file contents to the index
   branch     List, create, or delete branches
   checkout   Checkout a branch or paths to the working tree
   clone      Clone a repository into a new directory
   commit     Record changes to the repository
   push       Update remote refs along with associated objects
   remote     Manage set of tracked repositories

See 'git help <command>' for more information on a specific command.

"""
from subprocess import call

from docopt import docopt

if __name__ == "__main__":
    args = docopt(__doc__, version="git version 1.7.4.4", options_first=True)
    print("global arguments:")
    print(args)
    print("command arguments:")

    argv = [args["<command>"]] + args["<args>"]
    if args["<command>"] == "add":
        # In case subcommand is implemented as python module:
        import git_add

        print(docopt(git_add.__doc__, argv=argv))
    elif args["<command>"] == "branch":
        # In case subcommand is a script in some other programming language:
        exit(call(["python", "git_branch.py"] + argv))
    elif args["<command>"] in "checkout clone commit push remote".split():
        # For the rest we'll just keep DRY:
        exit(call(["python", "git_%s.py" % args["<command>"]] + argv))
    elif args["<command>"] in ["help", None]:
        exit(call(["python", "git.py", "--help"]))
    else:
        exit("%r is not a git.py command. See 'git help'." % args["<command>"])



"""Docopt is a Pythonic command-line interface parser that will make you smile.

Now: with spellcheck, flag extension (de-abbreviation), and capitalization fixes.
(but only when unambiguous)

 * Licensed under terms of MIT license (see LICENSE-MIT)

Contributors (roughly in chronological order):

 * Copyright (c) 2012 Andrew Kassen <atkassen@ucdavis.edu>
 * Copyright (c) 2012 jeffrimko <jeffrimko@gmail.com>
 * Copyright (c) 2012 Andrew Sutton <met48@met48.com>
 * Copyright (c) 2012 Andrew Sutton <met48@met48.com>
 * Copyright (c) 2012 Nima Johari <nimajohari@gmail.com>
 * Copyright (c) 2012-2013 Vladimir Keleshev, vladimir@keleshev.com
 * Copyright (c) 2014-2018 Matt Boersma <matt@sprout.org>
 * Copyright (c) 2016 amir <ladsgroup@gmail.com>
 * Copyright (c) 2015 Benjamin Bach <benjaoming@gmail.com>
 * Copyright (c) 2017 Oleg Bulkin <o.bulkin@gmail.com>
 * Copyright (c) 2018 Iain Barnett <iainspeed@gmail.com>
 * Copyright (c) 2019 itdaniher, itdaniher@gmail.com

"""
from __future__ import annotations

import re
import sys
from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

from ._version import __version__ as __version__

__all__ = ["docopt", "DocoptExit"]


def levenshtein_norm(source: str, target: str) -> float:
    """Calculates the normalized Levenshtein distance between two string
    arguments. The result will be a float in the range [0.0, 1.0], with 1.0
    signifying the biggest possible distance between strings with these lengths
    """

    # Compute Levenshtein distance using helper function. The max is always
    # just the length of the longer string, so this is used to normalize result
    # before returning it
    distance = levenshtein(source, target)
    return float(distance) / max(len(source), len(target))


def levenshtein(source: str, target: str) -> int:
    """Computes the Levenshtein
    (https://en.wikipedia.org/wiki/Levenshtein_distance)
    and restricted Damerau-Levenshtein
    (https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance)
    distances between two Unicode strings with given lengths using the
    Wagner-Fischer algorithm
    (https://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm).
    These distances are defined recursively, since the distance between two
    strings is just the cost of adjusting the last one or two characters plus
    the distance between the prefixes that exclude these characters (e.g. the
    distance between "tester" and "tested" is 1 + the distance between "teste"
    and "teste"). The Wagner-Fischer algorithm retains this idea but eliminates
    redundant computations by storing the distances between various prefixes in
    a matrix that is filled in iteratively.
    """

    # Create matrix of correct size (this is s_len + 1 * t_len + 1 so that the
    # empty prefixes "" can also be included). The leftmost column represents
    # transforming various source prefixes into an empty string, which can
    # always be done by deleting all characters in the respective prefix, and
    # the top row represents transforming the empty string into various target
    # prefixes, which can always be done by inserting every character in the
    # respective prefix. The ternary used to build the list should ensure that
    # this row and column are now filled correctly
    s_range = range(len(source) + 1)
    t_range = range(len(target) + 1)
    matrix = [[(i if j == 0 else j) for j in t_range] for i in s_range]

    # Iterate through rest of matrix, filling it in with Levenshtein
    # distances for the remaining prefix combinations
    for i in s_range[1:]:
        for j in t_range[1:]:
            # Applies the recursive logic outlined above using the values
            # stored in the matrix so far. The options for the last pair of
            # characters are deletion, insertion, and substitution, which
            # amount to dropping the source character, the target character,
            # or both and then calculating the distance for the resulting
            # prefix combo. If the characters at this point are the same, the
            # situation can be thought of as a free substitution
            del_dist = matrix[i - 1][j] + 1
            ins_dist = matrix[i][j - 1] + 1
            sub_trans_cost = 0 if source[i - 1] == target[j - 1] else 1
            sub_dist = matrix[i - 1][j - 1] + sub_trans_cost

            # Choose option that produces smallest distance
            matrix[i][j] = min(del_dist, ins_dist, sub_dist)

    # At this point, the matrix is full, and the biggest prefixes are just the
    # strings themselves, so this is the desired distance
    return matrix[len(source)][len(target)]


class DocoptLanguageError(Exception):

    """Error in construction of usage-message by developer."""


class DocoptExit(SystemExit):

    """Exit in case user invoked program with incorrect arguments."""

    usage = ""

    def __init__(
        self,
        message: str = "",
        collected: list[Pattern] | None = None,
        left: list[Pattern] | None = None,
    ) -> None:
        self.collected = collected if collected is not None else []
        self.left = left if left is not None else []
        SystemExit.__init__(self, (message + "\n" + self.usage).strip())


class Pattern:
    def __init__(
        self, name: str | None, value: list[str] | str | int | None = None
    ) -> None:
        self._name, self.value = name, value

    @property
    def name(self) -> str | None:
        return self._name

    def __eq__(self, other) -> bool:
        return repr(self) == repr(other)

    def __hash__(self) -> int:
        return hash(repr(self))


def transform(pattern: BranchPattern) -> Either:
    """Expand pattern into an (almost) equivalent one, but with single Either.

    Example: ((-a | -b) (-c | -d)) => (-a -c | -a -d | -b -c | -b -d)
    Quirks: [-a] => (-a), (-a...) => (-a -a)

    """
    result = []
    groups = [[pattern]]
    while groups:
        children = groups.pop(0)
        parents = [Required, NotRequired, OptionsShortcut, Either, OneOrMore]
        if any(t in map(type, children) for t in parents):
            child = [c for c in children if type(c) in parents][0]
            children.remove(child)
            if type(child) is Either:
                for c in child.children:
                    groups.append([c] + children)
            elif type(child) is OneOrMore:
                groups.append(child.children * 2 + children)
            else:
                groups.append(child.children + children)
        else:
            result.append(children)
    return Either(*[Required(*e) for e in result])


TSingleMatch = Tuple[Union[int, None], Union["LeafPattern", None]]


class LeafPattern(Pattern):

    """Leaf/terminal node of a pattern tree."""

    def __repr__(self) -> str:
        return "%s(%r, %r)" % (self.__class__.__name__, self.name, self.value)

    def single_match(self, left: list[LeafPattern]) -> TSingleMatch:
        raise NotImplementedError  # pragma: no cover

    def flat(self, *types) -> list[LeafPattern]:
        return [self] if not types or type(self) in types else []

    def match(
        self, left: list[LeafPattern], collected: list[Pattern] | None = None
    ) -> tuple[bool, list[LeafPattern], list[Pattern]]:
        collected = [] if collected is None else collected
        increment: Any | None = None
        pos, match = self.single_match(left)
        if match is None or pos is None:
            return False, left, collected
        left_ = left[:pos] + left[(pos + 1) :]
        same_name = [a for a in collected if a.name == self.name]
        if type(self.value) == int and len(same_name) > 0:
            if isinstance(same_name[0].value, int):
                same_name[0].value += 1
            return True, left_, collected
        if type(self.value) == int and not same_name:
            match.value = 1
            return True, left_, collected + [match]
        if same_name and type(self.value) == list:
            if type(match.value) == str:
                increment = [match.value]
            if same_name[0].value is not None and increment is not None:
                if isinstance(same_name[0].value, type(increment)):
                    same_name[0].value += increment
            return True, left_, collected
        elif not same_name and type(self.value) == list:
            if isinstance(match.value, str):
                match.value = [match.value]
            return True, left_, collected + [match]
        return True, left_, collected + [match]


class BranchPattern(Pattern):

    """Branch/inner node of a pattern tree."""

    def __init__(self, *children) -> None:
        self.children = list(children)

    def match(self, left: list[Pattern], collected: list[Pattern] | None = None) -> Any:
        raise NotImplementedError  # pragma: no cover

    def fix(self) -> "BranchPattern":
        self.fix_identities()
        self.fix_repeating_arguments()
        return self

    def fix_identities(self, uniq: Any | None = None) -> None:
        """Make pattern-tree tips point to same object if they are equal."""
        flattened = self.flat()
        uniq = list(set(flattened)) if uniq is None else uniq
        for i, child in enumerate(self.children):
            if not hasattr(child, "children"):
                assert child in uniq
                self.children[i] = uniq[uniq.index(child)]
            else:
                child.fix_identities(uniq)
        return None

    def fix_repeating_arguments(self) -> BranchPattern:
        """Fix elements that should accumulate/increment values."""
        either = [list(child.children) for child in transform(self).children]
        for case in either:
            for e in [child for child in case if case.count(child) > 1]:
                if type(e) is Argument or type(e) is Option and e.argcount:
                    if e.value is None:
                        e.value = []
                    elif type(e.value) is not list:
                        e.value = cast(str, e.value)
                        e.value = e.value.split()
                if type(e) is Command or type(e) is Option and e.argcount == 0:
                    e.value = 0
        return self

    def __repr__(self) -> str:
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(repr(a) for a in self.children),
        )

    def flat(self, *types) -> Any:
        if type(self) in types:
            return [self]
        return sum([child.flat(*types) for child in self.children], [])


class Argument(LeafPattern):
    def single_match(self, left: list[LeafPattern]) -> TSingleMatch:
        for n, pattern in enumerate(left):
            if type(pattern) is Argument:
                return n, Argument(self.name, pattern.value)
        return None, None


class Command(Argument):
    def __init__(self, name: str | None, value: bool = False) -> None:
        self._name, self.value = name, value

    def single_match(self, left: list[LeafPattern]) -> TSingleMatch:
        for n, pattern in enumerate(left):
            if type(pattern) is Argument:
                if pattern.value == self.name:
                    return n, Command(self.name, True)
                else:
                    break
        return None, None


class Option(LeafPattern):
    def __init__(
        self,
        short: str | None = None,
        longer: str | None = None,
        argcount: int = 0,
        value: list[str] | str | int | None = False,
    ) -> None:
        assert argcount in (0, 1)
        self.short, self.longer, self.argcount = short, longer, argcount
        self.value = None if value is False and argcount else value

    @classmethod
    def parse(class_, option_description: str) -> Option:
        short, longer, argcount, value = None, None, 0, False
        options, description = re.split(
            r"(?:  )|$", option_description.strip(), flags=re.M, maxsplit=1
        )
        options = options.replace(",", " ").replace("=", " ")
        for s in options.split():
            if s.startswith("--"):
                longer = s
            elif s.startswith("-"):
                short = s
            else:
                argcount = 1
        if argcount:
            matched = re.findall(r"\[default: (.*)\]", description, flags=re.I)
            value = matched[0] if matched else None
        return class_(short, longer, argcount, value)

    def single_match(self, left: list[LeafPattern]) -> TSingleMatch:
        for n, pattern in enumerate(left):
            if self.name == pattern.name:
                return n, pattern
        return None, None

    @property
    def name(self) -> str | None:
        return self.longer or self.short

    def __repr__(self) -> str:
        return "Option(%r, %r, %r, %r)" % (
            self.short,
            self.longer,
            self.argcount,
            self.value,
        )


class Required(BranchPattern):
    def match(self, left: list[Pattern], collected: list[Pattern] | None = None) -> Any:
        collected = [] if collected is None else collected
        original_collected = collected
        original_left = left
        for pattern in self.children:
            matched, left, collected = pattern.match(left, collected)
            if not matched:
                return False, original_left, original_collected
        return True, left, collected


class NotRequired(BranchPattern):
    def match(self, left: list[Pattern], collected: list[Pattern] | None = None) -> Any:
        collected = [] if collected is None else collected
        for pattern in self.children:
            _, left, collected = pattern.match(left, collected)
        return True, left, collected


class OptionsShortcut(NotRequired):

    """Marker/placeholder for [options] shortcut."""


class OneOrMore(BranchPattern):
    def match(self, left: list[Pattern], collected: list[Pattern] | None = None) -> Any:
        assert len(self.children) == 1
        collected = [] if collected is None else collected
        original_collected = collected
        original_left = left
        last_left = None
        matched = True
        times = 0
        while matched:
            matched, left, collected = self.children[0].match(left, collected)
            times += 1 if matched else 0
            if last_left == left:
                break
            last_left = left
        if times >= 1:
            return True, left, collected
        return False, original_left, original_collected


class Either(BranchPattern):
    def match(self, left: list[Pattern], collected: list[Pattern] | None = None) -> Any:
        collected = [] if collected is None else collected
        outcomes = []
        for pattern in self.children:
            matched, _, _ = outcome = pattern.match(left, collected)
            if matched:
                outcomes.append(outcome)
        if outcomes:
            return min(outcomes, key=lambda outcome: len(outcome[1]))
        return False, left, collected


class Tokens(list):
    def __init__(
        self,
        source: list[str] | str,
        error: Type[DocoptExit] | Type[DocoptLanguageError] = DocoptExit,
    ) -> None:
        if isinstance(source, list):
            self += source
        else:
            self += source.split()
        self.error = error

    @staticmethod
    def from_pattern(source: str) -> Tokens:
        source = re.sub(r"([\[\]\(\)\|]|\.\.\.)", r" \1 ", source)
        fragments = [s for s in re.split(r"\s+|(\S*<.*?>)", source) if s]
        return Tokens(fragments, error=DocoptLanguageError)

    def move(self) -> str | None:
        return self.pop(0) if len(self) else None

    def current(self) -> str | None:
        return self[0] if len(self) else None


def parse_longer(
    tokens: Tokens, options: list[Option], argv: bool = False, more_magic: bool = False
) -> list[Pattern]:
    """longer ::= '--' chars [ ( ' ' | '=' ) chars ] ;"""
    current_token = tokens.move()
    if current_token is None or not current_token.startswith("--"):
        raise ValueError(
            f"parse_longer got what appears to be an invalid token: {current_token}"
        )
    longer, maybe_eq, maybe_value = current_token.partition("=")
    if maybe_eq == maybe_value == "":
        value = None
    else:
        value = maybe_value
    similar = [o for o in options if o.longer and longer == o.longer]
    start_collision = (
        len(
            [
                o
                for o in options
                if o.longer and longer in o.longer and o.longer.startswith(longer)
            ]
        )
        > 1
    )
    if argv and not len(similar) and not start_collision:
        similar = [
            o
            for o in options
            if o.longer and longer in o.longer and o.longer.startswith(longer)
        ]
    # try advanced matching
    if more_magic and not similar:
        corrected = [
            (longer, o)
            for o in options
            if o.longer and levenshtein_norm(longer, o.longer) < 0.25
        ]
        if corrected:
            print(f"NB: Corrected {corrected[0][0]} to {corrected[0][1].longer}")
        similar = [correct for (original, correct) in corrected]
    if len(similar) > 1:
        raise DocoptLanguageError(f"{longer} is not a unique prefix: {similar}?")
    elif len(similar) < 1:
        argcount = 1 if maybe_eq == "=" else 0
        o = Option(None, longer, argcount)
        options.append(o)
        if tokens.error is DocoptExit:
            o = Option(None, longer, argcount, value if argcount else True)
    else:
        o = Option(
            similar[0].short, similar[0].longer, similar[0].argcount, similar[0].value
        )
        if o.argcount == 0:
            if value is not None:
                raise tokens.error("%s must not have an argument" % o.longer)
        else:
            if value is None:
                if tokens.current() in [None, "--"]:
                    raise tokens.error("%s requires argument" % o.longer)
                value = tokens.move()
        if tokens.error is DocoptExit:
            o.value = value if value is not None else True
    return [o]


def parse_shorts(
    tokens: Tokens, options: list[Option], more_magic: bool = False
) -> list[Pattern]:
    """shorts ::= '-' ( chars )* [ [ ' ' ] chars ] ;"""
    token = tokens.move()
    if token is None or not token.startswith("-") or token.startswith("--"):
        raise ValueError(
            f"parse_shorts got what appears to be an invalid token: {token}"
        )
    left = token.lstrip("-")
    parsed: list[Pattern] = []
    while left != "":
        short, left = "-" + left[0], left[1:]
        transformations: dict[str | None, Callable[[str], str]] = {None: lambda x: x}
        if more_magic:
            transformations["lowercase"] = lambda x: x.lower()
            transformations["uppercase"] = lambda x: x.upper()
        # try identity, lowercase, uppercase, iff such resolves uniquely
        # (ie if upper and lowercase are not both defined)
        similar: list[Option] = []
        de_abbreviated = False
        for transform_name, transform in transformations.items():
            transformed = list(set([transform(o.short) for o in options if o.short]))
            no_collisions = len(
                [
                    o
                    for o in options
                    if o.short and transformed.count(transform(o.short)) == 1
                ]
            )  # == len(transformed)
            if no_collisions:
                similar = [
                    o
                    for o in options
                    if o.short and transform(o.short) == transform(short)
                ]
                if similar:
                    if transform_name:
                        print(
                            f"NB: Corrected {short} to {similar[0].short} "
                            f"via {transform_name}"
                        )
                    break
            # if transformations do not resolve, try abbreviations of 'longer' forms
            # iff such resolves uniquely (ie if no two longer forms begin with the
            # same letter)
            if not similar and more_magic:
                abbreviated = [
                    transform(o.longer[1:3])
                    for o in options
                    if o.longer and not o.short
                ] + [transform(o.short) for o in options if o.short and not o.longer]
                nonredundantly_abbreviated_options = [
                    o for o in options if o.longer and abbreviated.count(short) == 1
                ]
                no_collisions = len(nonredundantly_abbreviated_options) == len(
                    abbreviated
                )
                if no_collisions:
                    for o in options:
                        if (
                            not o.short
                            and o.longer
                            and transform(short) == transform(o.longer[1:3])
                        ):
                            similar = [o]
                            print(
                                f"NB: Corrected {short} to {similar[0].longer} "
                                f"via abbreviation (case change: {transform_name})"
                            )
                            break
                if len(similar):
                    de_abbreviated = True
                    break
        if len(similar) > 1:
            raise DocoptLanguageError(
                f"{short} is specified ambiguously {len(similar)} times"
            )
        elif len(similar) < 1:
            o = Option(short, None, 0)
            options.append(o)
            if tokens.error is DocoptExit:
                o = Option(short, None, 0, True)
        else:
            if de_abbreviated:
                option_short_value = None
            else:
                option_short_value = transform(short)
            o = Option(
                option_short_value,
                similar[0].longer,
                similar[0].argcount,
                similar[0].value,
            )
            value = None
            current_token = tokens.current()
            if o.argcount != 0:
                if left == "":
                    if current_token is None or current_token == "--":
                        raise tokens.error("%s requires argument" % short)
                    else:
                        value = tokens.move()
                else:
                    value = left
                    left = ""
            if tokens.error is DocoptExit:
                o.value = value if value is not None else True
        parsed.append(o)
    return parsed


def parse_pattern(source: str, options: list[Option]) -> Required:
    tokens = Tokens.from_pattern(source)
    result = parse_expr(tokens, options)
    if tokens.current() is not None:
        raise tokens.error("unexpected ending: %r" % " ".join(tokens))
    return Required(*result)


def parse_expr(tokens: Tokens, options: list[Option]) -> list[Pattern]:
    """expr ::= seq ( '|' seq )* ;"""
    result: list[Pattern] = []
    seq_0: list[Pattern] = parse_seq(tokens, options)
    if tokens.current() != "|":
        return seq_0
    if len(seq_0) > 1:
        result.append(Required(*seq_0))
    else:
        result += seq_0
    while tokens.current() == "|":
        tokens.move()
        seq_1 = parse_seq(tokens, options)
        if len(seq_1) > 1:
            result += [Required(*seq_1)]
        else:
            result += seq_1
    return [Either(*result)]


def parse_seq(tokens: Tokens, options: list[Option]) -> list[Pattern]:
    """seq ::= ( atom [ '...' ] )* ;"""
    result: list[Pattern] = []
    while tokens.current() not in [None, "]", ")", "|"]:
        atom = parse_atom(tokens, options)
        if tokens.current() == "...":
            atom = [OneOrMore(*atom)]
            tokens.move()
        result += atom
    return result


def parse_atom(tokens: Tokens, options: list[Option]) -> list[Pattern]:
    """atom ::= '(' expr ')' | '[' expr ']' | 'options'
    | longer | shorts | argument | command ;
    """
    token = tokens.current()
    if not token:
        return [Command(tokens.move())]  # pragma: no cover
    elif token in "([":
        tokens.move()
        matching = {"(": ")", "[": "]"}[token]
        pattern = {"(": Required, "[": NotRequired}[token]
        matched_pattern = pattern(*parse_expr(tokens, options))
        if tokens.move() != matching:
            raise tokens.error("unmatched '%s'" % token)
        return [matched_pattern]
    elif token == "options":
        tokens.move()
        return [OptionsShortcut()]
    elif token.startswith("--") and token != "--":
        return parse_longer(tokens, options)
    elif token.startswith("-") and token not in ("-", "--"):
        return parse_shorts(tokens, options)
    elif token.startswith("<") and token.endswith(">") or token.isupper():
        return [Argument(tokens.move())]
    else:
        return [Command(tokens.move())]


def parse_argv(
    tokens: Tokens,
    options: list[Option],
    options_first: bool = False,
    more_magic: bool = False,
) -> list[Pattern]:
    """Parse command-line argument vector.

    If options_first:
        argv ::= [ longer | shorts ]* [ argument ]* [ '--' [ argument ]* ] ;
    else:
        argv ::= [ longer | shorts | argument ]* [ '--' [ argument ]* ] ;

    """

    def isanumber(x):
        try:
            float(x)
            return True
        except ValueError:
            return False

    parsed: list[Pattern] = []
    current_token = tokens.current()
    while current_token is not None:
        if current_token == "--":
            return parsed + [Argument(None, v) for v in tokens]
        elif current_token.startswith("--"):
            parsed += parse_longer(tokens, options, argv=True, more_magic=more_magic)
        elif (
            current_token.startswith("-")
            and current_token != "-"
            and not isanumber(current_token)
        ):
            parsed += parse_shorts(tokens, options, more_magic=more_magic)
        elif options_first:
            return parsed + [Argument(None, v) for v in tokens]
        else:
            parsed.append(Argument(None, tokens.move()))
        current_token = tokens.current()
    return parsed


class DocSections(NamedTuple):
    before_usage: str
    usage_header: str
    usage_body: str
    after_usage: str


def parse_docstring_sections(docstring: str) -> DocSections:
    """Partition the docstring into the main sections.

    The docstring is returned, split into a tuple of 4 pieces: text before the
    usage section, the usage section header, the usage section body and text
    following the usage section.
    """
    usage_pattern = r"""
    # Any number of lines (that don't include usage:) precede the usage section
    \A(?P<before_usage>(?:(?!.*\busage:).*\n)*)
    # The `usage:` section header.
    ^(?P<usage_header>.*\busage:)
    (?P<usage_body>
        # The first line of the body may follow the header without a line break:
        (?:.*(?:\n|\Z))
        # Any number of additional indented lines
        (?:[ \t].*(?:\n|\Z))*
    )
    # Everything else
    (?P<after_usage>(?:.|\n)*)\Z
    """
    match = re.match(usage_pattern, docstring, flags=re.M | re.I | re.VERBOSE)
    if not match:
        raise DocoptLanguageError(
            'Failed to parse doc: "usage:" section (case-insensitive) not found. '
            "Check http://docopt.org/ for examples of how your doc should look."
        )
    before, header, body, after = match.groups()
    return DocSections(before, header, body, after)


def parse_options(docstring: str) -> list[Option]:
    """Parse the option descriptions from the help text.

    `docstring` is the sub-section of the overall docstring that option
    descriptions should be parsed from. It must not contain the "usage:"
    section, as wrapped lines in the usage pattern can be misinterpreted as
    option descriptions.

    Option descriptions appear below the usage patterns, They define synonymous
    long and short options, options that have arguments, and the default values
    of options' arguments. They look like this:

    ```
        -v, --verbose             Be more verbose
        -n COUNT, --number COUNT  The number of times to
                                do the thing  [default: 42]
    ```
    """
    option_start = r"""
    # Option descriptions begin on a new line
    ^
    # They may occur on the same line as an options: section heading
    (?:.*options:)?
    # They can be indented with whitespace
    [ \t]*
    # The description itself starts with the short or long flag (-x or --xxx)
    (-\S)
    """
    parts = re.split(option_start, docstring, flags=re.M | re.I | re.VERBOSE)[1:]
    return [
        Option.parse(start + rest) for (start, rest) in zip(parts[0::2], parts[1::2])
    ]


def lint_docstring(sections: DocSections):
    """Report apparent mistakes in the docstring format."""
    if re.search("options:", sections.usage_body, flags=re.I):
        raise DocoptLanguageError(
            'Failed to parse docstring: "options:" (case-insensitive) was '
            'found in "usage:" section. Use a blank line after the usage, or '
            "start the next section without leading whitespace."
        )
    if re.search("usage:", sections.usage_body + sections.after_usage, flags=re.I):
        raise DocoptLanguageError(
            'Failed to parse docstring: More than one "usage:" '
            "(case-insensitive) section found."
        )
    if sections.usage_body.strip() == "":
        raise DocoptLanguageError(
            'Failed to parse docstring: "usage:" section is empty.'
            "Check http://docopt.org/ for examples of how your doc should look."
        )


def formal_usage(usage: str) -> str:
    program_name, *tokens = usage.split()
    return "( " + " ".join(") | (" if s == program_name else s for s in tokens) + " )"


def extras(
    default_help: bool, version: None, options: list[Pattern], docstring: str
) -> None:
    if default_help and any(
        (o.name in ("-h", "--help")) and o.value
        for o in options
        if isinstance(o, Option)
    ):
        print(docstring.strip("\n"))
        sys.exit()
    if version and any(
        o.name == "--version" and o.value for o in options if isinstance(o, Option)
    ):
        print(version)
        sys.exit()


class ParsedOptions(dict):
    def __repr__(self):
        return "{%s}" % ",\n ".join("%r: %r" % i for i in sorted(self.items()))

    def __getattr__(self, name: str) -> str | bool | None:
        return self.get(name) or {
            name: self.get(k)
            for k in self.keys()
            if name in [k.lstrip("-").replace("-", "_"), k.lstrip("<").rstrip(">")]
        }.get(name)


def docopt(
    docstring: str,
    argv: list[str] | str | None = None,
    default_help: bool = True,
    version: Any = None,
    options_first: bool = False,
) -> ParsedOptions:
    """Parse `argv` based on command-line interface described in `docstring`.

    `docopt` creates your command-line interface based on its
    description that you pass as `docstring`. Such description can contain
    --options, <positional-argument>, commands, which could be
    [optional], (required), (mutually | exclusive) or repeated...

    Parameters
    ----------
    docstring : str
        Description of your command-line interface.
    argv : list of str or str, optional
        Argument vector to be parsed. sys.argv[1:] is used if not
        provided. If str is passed, the string is split on whitespace.
    default_help : bool (default: True)
        Set to False to disable automatic help on -h or --help
        options.
    version : any object
        If passed, the object will be printed if --version is in
        `argv`.
    options_first : bool (default: False)
        Set to True to require options precede positional arguments,
        i.e. to forbid options and positional arguments intermix.

    Returns
    -------
    arguments: dict-like
        A dictionary, where keys are names of command-line elements
        such as e.g. "--verbose" and "<path>", and values are the
        parsed values of those elements. Also supports dot access.

    Example
    -------
    >>> from docopt import docopt
    >>> doc = '''
    ... Usage:
    ...     my_program tcp <host> <port> [--timeout=<seconds>]
    ...     my_program serial <port> [--baud=<n>] [--timeout=<seconds>]
    ...     my_program (-h | --help | --version)
    ...
    ... Options:
    ...     -h, --help  Show this screen and exit.
    ...     --baud=<n>  Baudrate [default: 9600]
    ... '''
    >>> argv = ['tcp', '127.0.0.1', '80', '--timeout', '30']
    >>> docopt(doc, argv)
    {'--baud': '9600',
     '--help': False,
     '--timeout': '30',
     '--version': False,
     '<host>': '127.0.0.1',
     '<port>': '80',
     'serial': False,
     'tcp': True}
    """
    argv = sys.argv[1:] if argv is None else argv
    sections = parse_docstring_sections(docstring)
    lint_docstring(sections)
    DocoptExit.usage = sections.usage_header + sections.usage_body
    options = [
        *parse_options(sections.before_usage),
        *parse_options(sections.after_usage),
    ]
    pattern = parse_pattern(formal_usage(sections.usage_body), options)
    pattern_options = set(pattern.flat(Option))
    for options_shortcut in pattern.flat(OptionsShortcut):
        options_shortcut.children = [
            opt for opt in options if opt not in pattern_options
        ]
    parsed_arg_vector = parse_argv(Tokens(argv), list(options), options_first)
    extras(default_help, version, parsed_arg_vector, docstring)
    matched, left, collected = pattern.fix().match(parsed_arg_vector)
    if matched and left == []:
        return ParsedOptions((a.name, a.value) for a in (pattern.flat() + collected))
    if left:
        raise DocoptExit(f"Warning: found unmatched (duplicate?) arguments {left}")
    raise DocoptExit(collected=collected, left=left)



"""Usage: prog.py [--count=N] PATH FILE...

Arguments:
  FILE     input file
  PATH     out directory

Options:
  --count=N   number of operations

"""
import os

from docopt import docopt

try:
    from schema import And
    from schema import Or
    from schema import Schema
    from schema import SchemaError
    from schema import Use
except ImportError:
    exit(
        "This example requires that `schema` data-validation library"
        " is installed: \n    pip install schema\n"
        "https://github.com/halst/schema"
    )


if __name__ == "__main__":
    args = docopt(__doc__)

    schema = Schema(
        {
            "FILE": [Use(open, error="FILE should be readable")],
            "PATH": And(os.path.exists, error="PATH should exist"),
            "--count": Or(
                None,
                And(Use(int), lambda n: 0 < n < 5),
                error="--count=N should be integer 0 < N < 5",
            ),
        }
    )
    try:
        args = schema.validate(args)
    except SchemaError as e:
        exit(e)

    print(args)



"""Usage: odd_even_example.py [-h | --help] (ODD EVEN)...

Example, try:
  odd_even_example.py 1 2 3 4

Options:
  -h, --help

"""
from docopt import docopt

if __name__ == "__main__":
    arguments = docopt(__doc__)
    print(arguments)



"""Usage: arguments_example.py [-vqrh] [FILE] ...
          arguments_example.py (--left | --right) CORRECTION FILE

Process FILE and optionally apply correction to either left-hand side or
right-hand side.

Arguments:
  FILE        optional input file
  CORRECTION  correction angle, needs FILE, --left or --right to be present

Options:
  -h --help
  -v       verbose mode
  -q       quiet mode
  -r       make report
  --left   use left-hand side
  --right  use right-hand side

"""
from docopt import docopt

if __name__ == "__main__":
    arguments = docopt(__doc__)
    print(arguments)



"""usage: git commit [options] [--] [<filepattern>...]

    -h, --help
    -q, --quiet           suppress summary after successful commit
    -v, --verbose         show diff in commit message template

Commit message options
    -F, --file <file>     read message from file
    --author <author>     override author for commit
    --date <date>         override date for commit
    -m, --message <message>
                          commit message
    -c, --reedit-message <commit>
                          reuse and edit message from specified commit
    -C, --reuse-message <commit>
                          reuse message from specified commit
    --fixup <commit>      use autosquash formatted message to fixup specified commit
    --squash <commit>     use autosquash formatted message to squash specified commit
    --reset-author        the commit is authored by me now
                          (used with -C-c/--amend)
    -s, --signoff         add Signed-off-by:
    -t, --template <file>
                          use specified template file
    -e, --edit            force edit of commit
    --cleanup <default>   how to strip spaces and #comments from message
    --status              include status in commit message template

Commit contents options
    -a, --all             commit all changed files
    -i, --include         add specified files to index for commit
    --interactive         interactively add files
    -o, --only            commit only specified files
    -n, --no-verify       bypass pre-commit hook
    --dry-run             show what would be committed
    --short               show status concisely
    --branch              show branch information
    --porcelain           machine-readable output
    -z, --null            terminate entries with NUL
    --amend               amend previous commit
    --no-post-rewrite     bypass post-rewrite hook
    -u, --untracked-files=<mode>
                          show untracked files, optional modes: all, normal, no.
                          [default: all]

"""
from docopt import docopt

if __name__ == "__main__":
    print(docopt(__doc__))



"""
usage: git remote [-v | --verbose]
       git remote add [-t <branch>] [-m <master>] [-f] [--mirror] <name> <url>
       git remote rename <old> <new>
       git remote rm <name>
       git remote set-head <name> (-a | -d | <branch>)
       git remote [-v | --verbose] show [-n] <name>
       git remote prune [-n | --dry-run] <name>
       git remote [-v | --verbose] update [-p | --prune] [(<group> | <remote>)...]
       git remote set-branches <name> [--add] <branch>...
       git remote set-url <name> <newurl> [<oldurl>]
       git remote set-url --add <name> <newurl>
       git remote set-url --delete <name> <url>

    -v, --verbose         be verbose; must be placed before a subcommand

"""
from docopt import docopt

if __name__ == "__main__":
    arguments = docopt(__doc__)
    print(arguments)



"""Usage:
  quick_example.py tcp <host> <port> [--timeout=<seconds>]
  quick_example.py serial <port> [--baud=9600] [--timeout=<seconds>]
  quick_example.py -h | --help | --version

"""
from docopt import docopt

if __name__ == "__main__":
    arguments = docopt(__doc__, version="0.1.1rc")
    print(arguments)



"""Not a serious example.

Usage:
  calculator_example.py <value> ( ( + | - | * | / ) <value> )...
  calculator_example.py <function> <value> [( , <value> )]...
  calculator_example.py (-h | --help)

Examples:
  calculator_example.py 1 + 2 + 3 + 4 + 5
  calculator_example.py 1 + 2 '*' 3 / 4 - 5    # note quotes around '*'
  calculator_example.py sum 10 , 20 , 30 , 40

Options:
  -h, --help

"""
from docopt import docopt

if __name__ == "__main__":
    arguments = docopt(__doc__)
    print(arguments)



"""
usage: git branch [options] [-r | -a] [--merged=<commit> | --no-merged=<commit>]
       git branch [options] [-l] [-f] <branchname> [<start-point>]
       git branch [options] [-r] (-d | -D) <branchname>
       git branch [options] (-m | -M) [<oldbranch>] <newbranch>

Generic options
    -h, --help
    -v, --verbose         show hash and subject, give twice for upstream branch
    -t, --track           set up tracking mode (see git-pull(1))
    --set-upstream        change upstream info
    --color=<when>        use colored output
    -r                    act on remote-tracking branches
    --contains=<commit>   print only branches that contain the commit
    --abbrev=<n>          use <n> digits to display SHA-1s

Specific git-branch actions:
    -a                    list both remote-tracking and local branches
    -d                    delete fully merged branch
    -D                    delete branch (even if not merged)
    -m                    move/rename a branch and its reflog
    -M                    move/rename a branch, even if target exists
    -l                    create the branch's reflog
    -f, --force           force creation (when already exists)
    --no-merged=<commit>  print only not merged branches
    --merged=<commit>     print only merged branches

"""
from docopt import docopt

if __name__ == "__main__":
    print(docopt(__doc__))



"""Example of program with many options using docopt.

Usage:
  options_example.py [-hvqrf NAME] [--exclude=PATTERNS]
                     [--select=ERRORS | --ignore=ERRORS] [--show-source]
                     [--statistics] [--count] [--benchmark] PATH...
  options_example.py (--doctest | --testsuite=DIR)
  options_example.py --version

Arguments:
  PATH  destination path

Options:
  -h --help            show this help message and exit
  --version            show version and exit
  -v --verbose         print status messages
  -q --quiet           report only file names
  -r --repeat          show all occurrences of the same error
  --exclude=PATTERNS   exclude files or directories which match these comma
                       separated patterns [default: .svn,CVS,.bzr,.hg,.git]
  -f NAME --file=NAME  when parsing directories, only check filenames matching
                       these comma separated patterns [default: *.py]
  --select=ERRORS      select errors and warnings (e.g. E,W6)
  --ignore=ERRORS      skip errors and warnings (e.g. E4,W)
  --show-source        show source code for each error
  --statistics         count errors and warnings
  --count              print total number of errors and warnings to standard
                       error and set exit code to 1 if total is not null
  --benchmark          measure processing speed
  --testsuite=DIR      run regression tests from dir
  --doctest            run doctest on myself

"""
from docopt import docopt

if __name__ == "__main__":
    arguments = docopt(__doc__, version="1.0.0rc2")
    print(arguments)
